'''
Created on Dec 26, 2017

@author: Salim
'''
from pylods.deserialize import Typed

def rename_attr(oldname, newname):
    def inner_rename_attr(cls):
        name = oldname
        if name.startswith("__"):
            name = "_" + cls.__name__ + oldname
        _create_pylods_property(cls, 'namedecode')
        _create_pylods_property(cls, 'nameencode')
        cls._pylods[cls]['nameencode'][name] = newname
        cls._pylods[cls]['namedecode'][newname] = name
        return cls
        
    return inner_rename_attr



def ignore_attr(name):
    def inner_ignore_attr(cls):
        _create_pylods_property(cls, 'ignore')
        cls._pylods[cls]['ignore'][name] = True
        return cls
    return inner_ignore_attr

def order_attr(name, order):
    def inner_order_attr(cls):
        _create_pylods_property(cls, 'order')
        cls._pylods[cls]['order'][name] = order
        return cls
    return inner_order_attr


def use_serializer(serializer):
    def inner_use_serializer(cls):
        _create_pylods_property(cls, 'serializer')
        cls._pylods[cls]['serializer'] = serializer()
        return cls
    return inner_use_serializer

def type_attr(name, typecls):
    def inner_type_attr(cls):
        Typed.register_type(name, typecls, cls)
        return cls
    return inner_type_attr

def use_deserializer(deserializer):
    def inner_use_deserializer(cls):
        _create_pylods_property(cls, 'deserializer')
        cls._pylods[cls]['deserializer'] = deserializer()
        return cls
    return inner_use_deserializer

def _create_pylods_property(cls, name):
    _create_pylods(cls)
    if not name in cls._pylods[cls]:
        cls._pylods[cls][name] = {}
        
def _create_pylods(cls):
    if(not hasattr(cls, "_pylods")):
        cls._pylods = {}
    if not cls in cls._pylods:
        cls._pylods[cls]={}





    



'''
Created on Nov 28, 2017

@author: Salim
'''
from abc import ABCMeta, abstractmethod


class Dictionary(object):
    
    __classmeta__ = ABCMeta
    
    
    @abstractmethod
    def gen_events(self, instream):
        '''
            generates events from parsing an input stream
        :param instream:
        '''
        
    
    @abstractmethod
    def is_obj_start(self, event):
        '''
            indicates if the given or a tuple representing an start of object event
            :param event:
        '''
    @abstractmethod
    def is_obj_end(self, event):
        '''
            indicates if the given or a tuple representing an end of object event
            :param event:
        '''
        
    @abstractmethod
    def is_value(self, event):
        '''
            indicates if the given or tuple representing an event that represents a raw value
            :param event:
        '''
    @abstractmethod    
    def is_obj_property_name(self, event):
        '''
         indicates if the given event or tuple representing an event is a property name. 
          This library expects to receive the value of the property next. If this value is 
          an object or array, start event for an array or map is expected.
          
        :param event:
        '''
         
    @abstractmethod
    def is_array_start(self, event):
        '''
            indicates if the given or tuple representing an event that indicates start of an array
            :param event:
        '''
    
    @abstractmethod
    def is_array_end(self, event):
        '''
            indicates if the given or tuple representing an event that indicates end of an array
            :param event:
        '''
        
    def read_value(self, event):
        '''
            Returns the value in the given event or tuple representing the event
        '''
    
    
    ######################### OBJECT
    
    @abstractmethod
    def write_object_start(self, numfields, outstream):
        raise Exception("Not Implemented!!!")
    
    @abstractmethod
    def write_object_end(self, numfields, outstream):
        raise Exception("Not Implemented!!!")
    
    @abstractmethod
    def write_object_field_separator(self, name, value, outstream):
        raise Exception("Not Implemented!!!")
    
    @abstractmethod
    def write_object_field_name(self, name, outstream):
        raise Exception("Not Implemented!!!")
    
    @abstractmethod
    def write_object_name_value_separator(self, name, value, outstream):
        raise Exception("Not Implemented!!!")

    ######################### ARRAY
    
    @abstractmethod 
    def write_array_start(self, length, outstream):
        raise Exception("Not Implemented!!!")
    
    @abstractmethod 
    def write_array_end(self, length, outstream):
        raise Exception("Not Implemented!!!")
    
    @abstractmethod 
    def write_array_field_separator(self, value, outstream):
        raise Exception("Not Implemented!!!")
     
    
    ######################### DICT
    
    @abstractmethod
    def write_dict_start(self, numfields, outstream):
        raise Exception("Not Implemented!!!")
    
    @abstractmethod
    def write_dict_end(self, numfields, outstream):
        raise Exception("Not Implemented!!!")
    
    @abstractmethod
    def write_dict_field_separator(self, name, value, outstream):
        raise Exception("Not Implemented!!!")    
    
    @abstractmethod
    def write_dict_field_name(self, name, outstream):
        raise Exception("Not Implemented!!!")
    
    @abstractmethod
    def write_dict_name_value_separator(self, name, value, outstream):
        raise Exception("Not Implemented!!!")

    
    @abstractmethod
    def write_value(self, val, outstream):
        raise Exception("Not Implemented!!!")    
    


'''
Created on Nov 26, 2017

@author: Salim
'''
from abc import ABCMeta, abstractmethod
from enum import Enum
from _pyio import __metaclass__
import types
from io import BytesIO
    
    
class DeserializationContext():
    
    def __init__(self):
        self.__attributes = {}
        
        
    def set_attribute(self, key, val):
        self.__attributes[key] = val

    def get_attribute(self, key, default=None):
        '''
            returns attribute from the attribute map. It will return None if it doens't exist
        :param key:
        '''
        return self.__attributes.get(key, default)
        
        
    def copy_all(self, ctxt):
        '''
         copies all values into the current context from given context
        :param ctxt:
        '''
        for key, value  in ctxt.iteritems:
            self.set_attribute(key, value)
        
        
    @classmethod 
    def create_context(cls, items=[]):
        ctxt = DeserializationContext()
        if items is not None:
            for key, value in items:
                ctxt.set_attribute(key, value)
        return ctxt
    
    
    
        
class Deserializer():
    '''
        Base Deserializer class. It only specifies that it needs a handled_class method
    '''
    __metaclass__ = ABCMeta
    
    @abstractmethod
    def execute(self, events, pdict, count, ctxt):
        raise Exception('Not implemented');
    
    
    

class EventBasedDeserializer(Deserializer):
    '''
        EventBasedDeserializer will execute deserializer over an iterator over events corresponding to the given object
    '''
    __metaclass__ = ABCMeta
    
    def execute(self, events, pdict, count, ctxt):
        tmp = EventStream(ClassEventIterator(events, pdict, count))
        val = self.deserialize(tmp, pdict, ctxt)
        try:
            while tmp.next():
                pass
        except StopIteration:
            pass
            
        return val
            
    
    @abstractmethod
    def deserialize(self, events, pdict, ctxt=None):
        raise Exception('Not implemented');
    
    
class ClassEventIterator(object):
    
    def __init__(self, events, pdict, count=0):
        self._events = events
        self._count = count
        self._pdict = pdict
        self._isdone = False
        if self._count == 0:
            if  self._pdict.is_obj_start(events.next()):
                self._count = self._count + 1
            else:
                raise Exception("Expected a object start event but didn't receive one")
    
    def __iter__(self):
        return self

    def __next__(self):
        if self._isdone:
            raise StopIteration()
        event = self._events.next()
        if event:
            if self._pdict.is_obj_start(event):
                self._count = self._count + 1
            elif self._pdict.is_obj_end(event):
                self._count = self._count - 1
                if self._count == 0:
                    self._isdone = True
                    raise StopIteration()
            return event
        else:
            return None
        
        

    next = __next__  # Python 2
    



        
        
    
class POPState(Enum):
    
    EXPECTING_OBJ_START = 1
    EXPECTING_OBJ_PROPERTY_OR_END = 3
    EXPECTING_VALUE = 4
    EXPECTING_ARRAY_START = 5
    EXPECTING_VALUE_OR_ARRAY_END = 6
    


class Typed():
    '''
        Basic class for classes that will support typed object based serialization. 
        Extending this class is not required to support type resolution however, 
        it will make the code cleaner
    '''
    
    _types = {}
        
        
    @classmethod
    def resolve(cls, propname, objcls=None):
        '''
            resolve type of the class property for the class. If objcls is not set then 
            assume that cls argument (class that the function is called from) is the class 
            we are trying to resolve the type for 
        :param cls:
        :param objcls:
        :param propname:
        '''
        # object class is not given
        if objcls is None:
            objcls = cls
        # get type
        typemap = cls._types.get(objcls)
        if(typemap is not None):
            # nothing registered for this property
            return typemap.get(propname, None)
        # type map didn't exist so none
        return None
    
    @classmethod
    def register_type(cls, propname, valcls, objcls=None):
        '''
            register type for class 
             
        :param cls:
        :param propname:
        :param valcls:
        :param objcls:
        '''
        # object class is not given
        if objcls is None:
            objcls = cls
        # get type map for objcls
        typemap = cls._types.get(objcls, None)
        # if type map doesn't exist then remove it
        if not typemap:
            typemap = {}
            cls._types[objcls] = typemap
        # register value class for property
        typemap[propname] = valcls
        
        





class Parser():
    '''
        Used for parsing to events
    '''
    
    __slots__ = ['_pdict']
    
    
    def __init__(self, pdict):
        self._pdict = pdict
        
        
    
    def parse(self, data):
        
        
        if type(data) is types.StringType:
            data = BytesIO(data);
            data.seek(0)
        if hasattr(data, 'read'):
            pass
        else:
            raise Exception('input should be of one the types [str, IOBase, file] but ' + str(type(data)) + " was received!!")
        
        return self._gen_events(data)
        
    def _gen_events(self, instream):
        '''
            generates events using the specific parser
        :param instream:
        '''
        return EventStream(self._pdict.gen_events(instream))
        
    
class EventStream():
    '''
        Wrapper Iterator as output of Parser
    '''
    
    def __init__(self, events):
        self._events = events
        
    
    def __iter__(self):
        return self

    def __next__(self):
        return  self._events.next()
    
    next = __next__  # Python 2
    
    






'''
Created on Nov 26, 2017

@author: Salim
'''



class ParseException(Exception):
    '''
        Exception to FormatTypeNotAcceptable
    '''
    def __init__(self, msg):
        Exception.__init__(self, msg)
        
        
class ObjectDeserializationException(ParseException):
    '''
        Exception to FormatTypeNotAcceptable
    '''
    def __init__(self, msg):
        Exception.__init__(self, msg)
        
        
class UnexpectedStateException(ParseException):
    '''
        Exception to FormatTypeNotAcceptable
    '''
    def __init__(self, expected, state=None, msg=None):
        txt = "Parser originally expected " + str(expected)
        if state is not None:
            txt = txt + " but state was " + str(state)
        
        if msg is not None:
            txt = txt + " with message: " + str(msg);
        Exception.__init__(self, txt)




'''
Created on Dec 26, 2017

@author: Salim
'''
from pylods.deserialize import  Typed, POPState, DeserializationContext
from pylods.error import UnexpectedStateException, \
    ObjectDeserializationException, ParseException
from abc import ABCMeta
from pylods.serialize import DataFormatGenerator


    
        
class Module():
    '''
        A Module contains customer serializer/deserializers.
    '''
    
    __metaclass__ = ABCMeta
    __slots__ = ['_serializers', '_deserializers']
    
    def __init__(self):
        self._serializers = {}
        self._deserializers = {}
    
    def add_serializer(self, cls, serializer):
        self._serializers[cls ] = serializer
    
    def add_deserializer(self, cls, deserializer):
        self._deserializers[cls ] = deserializer
    
    
    def get_serializers(self):
        return self._serializers
    
    def get_deserializers(self):
        return self._deserializers
    
    serializers = property(get_serializers)
    deserializers = property(get_deserializers)



class DecoratorsModule(Module):
    '''
        Default module for decorators    
    '''
    
    all_serializers = {}
    all_deserializers = {}
    
    def __init__(self):
        self._serializers = self.all_serializers
        self._deserializers = self.all_deserializers
        
    @classmethod
    def register_serializer(cls, scls, serializer):
        cls.all_serializers[scls ] = serializer();
    
    @classmethod
    def register_deserializer(cls, dcls, deserializer):
        cls.all_deserializers[dcls ] = deserializer();


class ObjectMapper(DataFormatGenerator):
    '''
        Main Parser class to deserialize to objects, values and etc. This parser requires a dictionary to serialize/ deserialize from an input stream
    '''
    
    __classmeta__ = ABCMeta
    
    __slots__ = [ '__deserializers']
    
    
    def __init__(self, pdict):
        super(ObjectMapper, self).__init__(pdict) 
        self.__deserializers = {}
#         events = self.prepare_input(events)
        self.register_module(DecoratorsModule())
        
        
    
#     def prepare_input(self, events):
#         if events is None:
#             return None   
#         if isinstance(events, EventStream):
#             return events
#         else:
#             raise UnsupportedTypeException('Input can only be an ' + str(EventStream) + ' instance!!')
    
    def read_value(self, events):
        val = self._pdict.read_value(events.next())
        return val
            

    def read_obj_property_name(self, events):
        propname = self._pdict.read_value(events.next());
        return propname
    
        
    def read_obj(self, events, cls=dict, state=POPState.EXPECTING_OBJ_START, ctxt=DeserializationContext.create_context()):
        cnt = 0
        if state is POPState.EXPECTING_OBJ_START:
            cnt = 0;
        elif state is POPState.EXPECTING_OBJ_PROPERTY_OR_END:
            cnt = 1
        else:
            raise ParseException("Couldn't start reading an object at this state: " + str(state))
        deserializer = self.__lookup_deserializer(cls)
        if deserializer:
            val = deserializer.execute(events, self._pdict, count=cnt, ctxt=ctxt)
        else:    
            val = self._read_obj(events, cls, state, ctxt)
        return val    
    
    
        
    def _read_obj(self, events, cls=dict, state=POPState.EXPECTING_OBJ_START, ctxt=DeserializationContext.create_context()):
        try:
            obj = cls()
        except:
            raise Exception("Failed to instantiate cls" + str(cls))
        if state is POPState.EXPECTING_OBJ_START:
            # setting to object start state
            first = events.next();
            # check if input is object
            if not self._pdict.is_obj_start(first):
                raise ObjectDeserializationException("Expected the content to start with an object!! But first event was: " + str(first))
        
        state = POPState.EXPECTING_OBJ_PROPERTY_OR_END
        propname = None
        event = events.next()
        while event:
            if state is POPState.EXPECTING_OBJ_PROPERTY_OR_END:
                # end of object
                if self._pdict.is_obj_end(event):
                    # check valid state
                    if state is not POPState.EXPECTING_OBJ_PROPERTY_OR_END:
                        raise UnexpectedStateException(state, POPState.EXPECTING_OBJ_PROPERTY_OR_END, " wasn't expecting a end of object")
                    return  obj
                elif self._pdict.is_obj_property_name(event):
                    # check valid state
                    if state is not POPState.EXPECTING_OBJ_PROPERTY_OR_END:
                        raise UnexpectedStateException(state, POPState.EXPECTING_OBJ_PROPERTY_OR_END, " the received object property name was not expected")
                    propname = self._pdict.read_value(event);
                    propname = self._decode_field_name(obj, propname)
                    # setting next state
                    state = POPState.EXPECTING_VALUE
                else:
                    raise ParseException(" Unexpected event when expected state is: " + str(state) + " while event was: " + str(event))
            elif state is POPState.EXPECTING_VALUE:
                val = None
                if self._pdict.is_array_start(event):
                    val = self.read_array(events, POPState.EXPECTING_VALUE_OR_ARRAY_END, cls, propname, ctxt)
                elif self._pdict.is_value(event):
                    # check valid state
                    if state is not POPState.EXPECTING_VALUE:
                        raise UnexpectedStateException(state, POPState.EXPECTING_VALUE, " wasn't expecting a value")
                    # read value
                    val = self._pdict.read_value(event)
                elif self._pdict.is_obj_start(event):
                    val = self._read_obj_as_value(events, cls, propname, ctxt)
                else:
                    raise ParseException('unrecognized event when reading value: ' + str(event))
                # setting value
                if cls is dict:
                    obj[propname] = val
                else:
                    setattr(obj, propname, val)
                propname = None
                state = POPState.EXPECTING_OBJ_PROPERTY_OR_END
            
            # continue
#             try:
            event = events.next()
#             except StopIteration:
#                 event = None
            
        raise UnexpectedStateException(state, POPState.EX, " wasn't expecting a value")
            
    def read_array(self, events, state=POPState.EXPECTING_ARRAY_START, cls=None, propname=None, ctxt=DeserializationContext.create_context()):
        res = []
        
        if state is POPState.EXPECTING_ARRAY_START:
            # setting to object start state
            first = events.next();
            # check if input is object
            if not self._pdict.is_array_start(first):
                raise UnexpectedStateException(POPState.EXPECTING_ARRAY_START, " didn't received start of array!!!")
        
        state = POPState.EXPECTING_VALUE_OR_ARRAY_END    
        event = events.next()
        while event:
            # end of object
            if self._pdict.is_array_end(event):
                # check valid state
                if state is not POPState.EXPECTING_VALUE_OR_ARRAY_END:
                    raise UnexpectedStateException(POPState.EXPECTING_VALUE_OR_ARRAY_END, state, " wasn't expecting a end of object")
                return res
            elif self._pdict.is_value(event):
                # check valid state
                if state is not POPState.EXPECTING_VALUE_OR_ARRAY_END:
                    raise UnexpectedStateException(state, POPState.EXPECTING_VALUE_OR_ARRAY_END, " wasn't expecting a value")
                val = self._pdict.read_value(event)
                res.append(val)
            elif self._pdict.is_obj_start(event):
                val = self._read_obj_as_value(events, cls, propname, ctxt)
                res.append(val)
            elif self._pdict.is_array_start(event):
                val = self.read_array(events, POPState.EXPECTING_VALUE_OR_ARRAY_END, cls, propname, ctxt)
                res.append(val)
            else:
                raise ParseException('Unexpected event')
            
            event = events.next()

    

    def _read_obj_as_value(self, events, cls=None, valname=None, ctxt=DeserializationContext.create_context()):
        valcls = None
        if cls is not None and valname is not None:
            # attempt to resolve class of value
            valcls = Typed.resolve(valname, cls)
        if cls is not None and valname is None:
            valcls = cls
        if valcls is None:
            valcls = dict
        deserializer = self.__lookup_deserializer(valcls)
        if deserializer:
            val = deserializer.execute(events, self._pdict, count=1, ctxt=ctxt)
        else:    
            val = self._read_obj(events, valcls, POPState.EXPECTING_OBJ_PROPERTY_OR_END, ctxt)
        return val
     
    
    def __lookup_deserializer(self, cls):
        deserializer = self.__deserializers.get(cls, None)
        if deserializer is None and hasattr(cls, '_pylods'):
            return cls._pylods[cls].get('deserializer', None)
        
        return None
    
    
    def register_module(self, module):
        for serializer in module.serializers.items():
            self.register_serializer(serializer[0], serializer[1])
            
        for deserializer in module.deserializers.items():
            self.register_deserializer(deserializer[0], deserializer[1])
        
    
    def register_deserializer(self, cls, deserializer):
        self.__deserializers[cls] = deserializer
        
        
    def copy(self):
        '''
         makes a clone copy of the mapper. It won't clone the serializers or deserializers and it won't copy the events
        '''
        try:
            tmp = self.__class__()
        except Exception:
            tmp = self.__class__(self._pdict)
            
        tmp._serializers = self._serializers
        tmp.__deserializers = self.__deserializers

        return tmp
    
    
        
        



'''
Created on Dec 5, 2017

@author: Salim
'''
from abc import abstractmethod, ABCMeta


class Serializer():
    '''
        Base Serializer class. It only specifies that it needs a handled_class method
    '''
    
    __metaclass__ = ABCMeta
    
    @abstractmethod
    def serialize(self, gen, obj, outstream):
        '''
        Is called by the serializer when trying to serialize this class
        :param gen:
        :param outstream:
        '''
        raise Exception('Not implemented');
    
    


class DataFormatGenerator():
    '''
        class that helps writing 
    '''
    
    __metaclass__ = ABCMeta
    
    __slots__ = ['_pdict', '_serializers']
    
    
    def __init__(self, pdict):
        self._pdict = pdict
        self._serializers = {}
    
    
    def write(self, val, outstream):
        if(self._isobject(val)):
            self._write_object(val, outstream)
        elif(self._isdict(val)):
            self._write_dict(val, outstream)
        elif(self._isarray(val)):
            self._write_array(val, outstream)
        else:
            self._write_value(val, outstream)
     
    def _isobject(self, val):
        return hasattr(val, '__dict__')  or hasattr(val, '__slots__')
    
    def _isdict(self, val):
        return isinstance(val, dict)
    
    def _isarray(self, val):
        return isinstance(val, list) or isinstance(val, tuple)
    
    def _write_value(self, val, outstream):
        self._pdict.write_value(val, outstream)
    
    
    def _write_array(self, val, outstream):
        length = len(val)
        self.write_array_start(length, outstream)
        if length > 0:
            self.write(val[0], outstream)
            idx = 1
            while idx < length:
                self.write_array_field_separator(val[idx], outstream)
                self.write(val[idx], outstream)
                idx += 1
        self.write_array_end(length, outstream)

    def _write_object(self, val, outstream):
        '''
        Write object into the stream
        :param val:
        :param outstream:
        '''
        # first attempt using an existing custom serializer
        serializer = self.__lookup_serializer(getattr(val, '__class__'))
        if serializer is not None:
            serializer.serialize(self, val, outstream)
            return
        
        fields = self._fetch_obj_fields(val)
        length = len(fields)
        self.write_object_start(length, outstream)
        count = 0;
        if length > 0:
            mappendname = self._encode_field_name(val, fields[0])
            if mappendname is not None:
                count += 1
                self.write_object_field(mappendname, getattr(val, fields[0]), outstream)
            idx = 1
            while idx < length:                
                name = fields[idx]
                mappedname = self._encode_field_name(val, name)
                if mappedname is not None:
                    value = getattr(val, name)
                    if count > 0:
                        self.write_object_field_separator(mappedname, value, outstream)
                    self.write_object_field(mappedname, value, outstream)
                    count += 1
                idx += 1
        self.write_object_end(length, outstream)
        
    def write_object_field(self, name, value, outstream):
        self.write_object_field_name(name, outstream)
        self.write_object_name_value_separator(name, value, outstream)
        self.write(value, outstream)
        
        
    def _write_dict(self, val, outstream):
        '''
        write a dictionary
        :param val:
        :param outstream:
        '''
        fields = val.keys()
        length = len(fields)
        self.write_dict_start(length, outstream)
        if length > 0:
            self.write_dict_field(fields[0], val[fields[0]], outstream)
            idx = 1
            while idx < length:
                name = fields[idx]
                value = val[fields[idx]]
                self.write_dict_field_separator(name, value, outstream)
                self.write_dict_field(name, value, outstream)
                idx += 1
        self.write_dict_end(length, outstream)        
    
    
    def write_dict_field(self, name, value, outstream):
        self.write_dict_field_name(name, outstream)
        self.write_dict_name_value_separator(name, value, outstream)
        self.write(value, outstream)
    
    def _fetch_obj_fields(self, obj):
        
        if  hasattr(obj, '__dict__'):
            fields = [f for f in obj.__dict__.keys() if not callable(obj.__dict__[f])]
        elif hasattr(obj, '__slots__'):
            fields = [f for f in obj.__slots__.keys() if not callable(obj.__slots__[f])]
        else:
            return [];
        
        fields = self._sort_obj_fields(obj, fields)
        return fields
    
    def _sort_obj_fields(self, obj, fields):
        cls = getattr(obj, '__class__')
        if hasattr(cls, '_pylods') and  cls in cls._pylods and  'order' in cls._pylods[cls]:
            tmp = self.__attach_order(cls, fields)
            tmp = sorted(tmp, key=lambda item: item[1])
            fields = [i[0] for i in tmp]
        return fields
    
    
    def __attach_order(self, cls, fields):
        attached = [None] * len(fields)
        for i in range(len(fields)):
            name = fields[i]
            if name in cls._pylods[cls]['order']:
                attached[i] = (name,cls._pylods[cls]['order'][name])
            else:
                attached[i] = (name,9999999999)     
        return attached           
    
    def _encode_field_name(self, obj, name):
        mappedname = name;
        
        # 1. check to rename fields using decorator
#         print(obj)
        cls = getattr(obj,"__class__")
        if hasattr(obj, '_pylods') and cls in cls._pylods :
            if 'ignore' in obj.__class__._pylods[cls] and name in obj._pylods[cls]['ignore']:
                return None 
            if 'nameencode' in obj.__class__._pylods[cls] and name in obj._pylods[cls]['nameencode']:
                return obj._pylods[cls]['nameencode'][name]
        # 2. check to ignore fields based on decorators
        # 3. remove private convention _
        if mappedname.startswith('_'):
            mappedname = mappedname[1:]
        return mappedname
    
    def _decode_field_name(self, obj, name):
        if isinstance(obj, dict):
            return name
        # 1. check to rename fields using decorator
        cls = getattr(obj,"__class__")
        if hasattr(obj, '_pylods'):            
            if 'namedecode' in obj.__class__._pylods[cls] and name in obj._pylods[cls]['namedecode']:
                return obj._pylods[cls]['namedecode'][name]
        # 2. check to ignore fields based on decorators
        # 3. remove private convention _
        fields = self._fetch_obj_fields(obj)
        properties=[p for p in dir(getattr(obj,"__class__")) if isinstance(getattr(getattr(obj,"__class__"),p),property)]
        if name in fields or name in properties:
            return name
        elif "_" + name in fields or properties:
            return "_" + name
        else:
            raise Exception("property \"" + str(name) + "\" couldn't be mapped to a property in object " + str(getattr(obj,"__class__")))
        
    ######################### ARRAY
    
    def write_array_start(self, length, outstream):
        self._pdict.write_array_start(length, outstream)
    
    def write_array_end(self, length, outstream):
        self._pdict.write_array_end(length, outstream)
    
    def write_array_field_separator(self, value, outstream):
        self._pdict.write_array_field_separator(value, outstream)
    
    
    ######################### OBJECT
     
    def write_object_start(self, numfields, outstream):
        self._pdict.write_object_start(numfields, outstream)
    
    def write_object_end(self, numfields, outstream):
        self._pdict.write_object_end(numfields, outstream)
    
    def write_object_field_separator(self, name, value, outstream):
        self._pdict.write_object_field_separator(name, value, outstream)
    
    def write_object_name_value_separator(self, name, value, outstream):
        self._pdict.write_object_name_value_separator(name, value, outstream)
        
    def write_object_field_name(self, name, outstream):
        self._pdict.write_object_field_name(name, outstream)
        
    ######################### DICT    
        
    def write_dict_start(self, numfields, outstream):
        self._pdict.write_dict_start(numfields, outstream)
    
    def write_dict_end(self, numfields, outstream):
        self._pdict.write_dict_end(numfields, outstream)
    
    def write_dict_field_separator(self, name, value, outstream):
        self._pdict.write_dict_field_separator(name, value, outstream)
        
    def write_dict_name_value_separator(self, name, value, outstream):
        self._pdict.write_dict_name_value_separator(name, value, outstream)    
    
    def write_dict_field_name(self, name, outstream):
        self._pdict.write_dict_field_name(name, outstream)
        
        
    def register_serializer(self, cls, serializer):
        self._serializers[cls] = serializer


    def __lookup_serializer(self, cls):
        deserializer = self._serializers.get(cls, None)
        if deserializer is None and hasattr(cls, '_pylods'):
            return cls._pylods[cls].get('serializer',None)
        
        return None

    
    
    

    
     
